// main.cpp
#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>
#include <JGL/JGL.h>
#include <Color4.hpp>
#include <memory>
#include <vector>
#include <string>
#include "ReNoise/perlin.hpp"     // Contains generatePerlinNoise1D, smoothNoise1D, generatePerlinNoise2D, smoothNoise2D
#include "ReUnirand/unirand.hpp"  

#include <J3ML/LinearAlgebra/Vector2.hpp>
using J3ML::LinearAlgebra::Vector2;

#define WIDTH 800
#define HEIGHT 600
#define DOT_SIZE 5

/**
 * @brief Maps a noise value in [-1, 1] to a y-coordinate on the window.
 *        +1 is at the top, -1 at the bottom.
 */
float mapNoiseToY(float noiseValue) {
    return HEIGHT - ((noiseValue + 1.0f) / 2.0f * HEIGHT);
}

/**
 * @brief A window class that displays Perlin noise in three modes.
 *
 * Mode 0: 1D graph - raw noise (red) and smoothed noise (blue) drawn as connected dots.
 * Mode 1: 2D raw noise image - the noise field is rendered as a grayscale image.
 * Mode 2: 2D smoothed noise image - the smoothed noise field is rendered as a grayscale image.
 *
 * Pressing the space bar or left mouse button cycles through the modes.
 */
class PerlinNoiseWindow : public ReWindow::OpenGLWindow {
public:
    PerlinNoiseWindow(const std::string &title, int width, int height)
        : ReWindow::OpenGLWindow(title, width, height, 2, 1),
          renderMode(0), toggleLocked(false),
          noise2DWidth(256), noise2DHeight(256)
    {
    }

    void OnOpen() override {
        if (!JGL::Init({ WIDTH, HEIGHT }, 0, 0))
            exit(0);

        // Create a render target with a white background.
        render_target = new RenderTarget({ WIDTH, HEIGHT }, Colors::White);

        // Generate 1D noise arrays.
        const int noiseResolution = 256;
        {
            ReUnirand::MarsagliaUniRng rng1;
            try {
                rng1.rinit(170);
            } catch (const std::exception &e) {
                ReWindow::Logger::Error("Error initializing RNG (1D): " + std::string(e.what()));
                exit(0);
            }
            rawNoise = ReNoise::generatePerlinNoise1D(noiseResolution, rng1);
            smoothedNoise1D = ReNoise::smoothNoise1D(rawNoise, 3);
            // Precompute drawing points for 1D noise.
            rawPoints.clear();
            smoothPoints.clear();
            float xSpacing = static_cast<float>(WIDTH) / (noiseResolution - 1);
            for (int i = 0; i < noiseResolution; ++i) {
                float x = i * xSpacing;
                float yRaw = mapNoiseToY(rawNoise[i]);
                float ySmooth = mapNoiseToY(smoothedNoise1D[i]);
                rawPoints.push_back(Vector2(x, yRaw));
                smoothPoints.push_back(Vector2(x, ySmooth));
            }
        }

        // Generate 2D noise arrays.
        {
            ReUnirand::MarsagliaUniRng rng2;
            try {
                rng2.rinit(170);
            } catch (const std::exception &e) {
                ReWindow::Logger::Error("Error initializing RNG (2D): " + std::string(e.what()));
                exit(0);
            }
            noise2D = ReNoise::generatePerlinNoise2D(noise2DWidth, noise2DHeight, rng2);
            noise2D_smoothed = ReNoise::smoothNoise2D(noise2D, noise2DWidth, noise2DHeight, 3);
        }
    }

    void OnRefresh(float elapsed) override {
        handleInput();
        JGL::Update({ WIDTH, HEIGHT });
        J2D::Begin(render_target, true);
            // Render based on the current mode.
            switch (renderMode) {
                case 0:
                    render1D();
                    break;
                case 1:
                    render2DRaw();
                    break;
                case 2:
                    render2DSmoothed();
                    break;
                default:
                    render1D();
                    break;
            }
        J2D::End();

        J2D::Begin(nullptr, true);
            J2D::DrawRenderTarget(render_target, { 0, 0 });
        J2D::End();

        SwapBuffers();
        PollEvents();
    }

    ~PerlinNoiseWindow() override {
        delete render_target;
    }

private:
    RenderTarget* render_target = nullptr;
    // 1D noise data:
    std::vector<float> rawNoise;
    std::vector<float> smoothedNoise1D;
    std::vector<Vector2> rawPoints;
    std::vector<Vector2> smoothPoints;
    // 2D noise data:
    std::vector<float> noise2D;
    std::vector<float> noise2D_smoothed;
    int noise2DWidth, noise2DHeight;

    // Render mode: 0 = 1D graph, 1 = 2D raw, 2 = 2D smoothed.
    int renderMode;
    bool toggleLocked;

    /**
     * @brief Renders the 1D noise graphs as connected dots.
     */
    void render1D() {
        // Draw raw 1D noise in red.
        for (size_t i = 0; i < rawPoints.size() - 1; ++i) {
            J2D::DrawLine(Colors::Red, rawPoints[i], rawPoints[i + 1], 2.0f);
        }
        for (const auto &pt : rawPoints) {
            J2D::DrawPoint(Colors::Red, pt, DOT_SIZE);
        }
        // Draw smoothed 1D noise in blue.
        for (size_t i = 0; i < smoothPoints.size() - 1; ++i) {
            J2D::DrawLine(Colors::Blue, smoothPoints[i], smoothPoints[i + 1], 2.0f);
        }
        for (const auto &pt : smoothPoints) {
            J2D::DrawPoint(Colors::Blue, pt, DOT_SIZE);
        }
    }

    /**
     * @brief Renders the 2D raw noise field as a grayscale image.
     */
    void render2DRaw() {
        float cellWidth = static_cast<float>(WIDTH) / noise2DWidth;
        float cellHeight = static_cast<float>(HEIGHT) / noise2DHeight;
        for (int j = 0; j < noise2DHeight; ++j) {
            for (int i = 0; i < noise2DWidth; ++i) {
                float value = noise2D[j * noise2DWidth + i];
                unsigned char gray = static_cast<unsigned char>(((value + 1.0f) / 2.0f) * 255);
                Color4 cellColor(gray, gray, gray, 255);
                Vector2 pos(i * cellWidth, j * cellHeight);
                Vector2 size(cellWidth, cellHeight);
                J2D::FillRect(cellColor, pos, size);
            }
        }
    }

    /**
     * @brief Renders the 2D smoothed noise field as a grayscale image.
     */
    void render2DSmoothed() {
        float cellWidth = static_cast<float>(WIDTH) / noise2DWidth;
        float cellHeight = static_cast<float>(HEIGHT) / noise2DHeight;
        for (int j = 0; j < noise2DHeight; ++j) {
            for (int i = 0; i < noise2DWidth; ++i) {
                float value = noise2D_smoothed[j * noise2DWidth + i];
                unsigned char gray = static_cast<unsigned char>(((value + 1.0f) / 2.0f) * 255);
                Color4 cellColor(gray, gray, gray, 255);
                Vector2 pos(i * cellWidth, j * cellHeight);
                Vector2 size(cellWidth, cellHeight);
                J2D::FillRect(cellColor, pos, size);
            }
        }
    }

    /**
     * @brief Handles keyboard and mouse input to cycle through render modes.
     *
     * Pressing the space bar or left mouse button cycles the renderMode value.
     * A simple lock prevents repeated toggling while the key/button is held down.
     */
    void handleInput() {
        if (IsKeyDown(Keys::Space) || IsMouseButtonDown(MouseButtons::Left)) {
            if (!toggleLocked) {
                renderMode = (renderMode + 1) % 3;
                toggleLocked = true;
            }
        } else {
            toggleLocked = false;
        }
        if (IsKeyDown(Keys::Escape)) {
            Close();
        }
    }
};

int main() {
    auto window = std::make_unique<PerlinNoiseWindow>("Perlin Noise Render", WIDTH, HEIGHT);
    if (!window->Open())
        return -1;

    while (!window->IsClosing()) {
        window->ManagedRefresh();
    }
    return 0;
}
