#include "ReNoise/perlin.hpp"
#include "ReUnirand/unirand.hpp" 
#include <cmath>
#include <vector>
#include <algorithm>

namespace ReNoise {

    float fade(float t) {
        // Ken Perlin's fade function: 6t^5 - 15t^4 + 10t^3.
        return t * t * t * (t * (t * 6 - 15) + 10);
    }

    float lerp(float a, float b, float t) {
        return a + t * (b - a);
    }

    std::vector<float> generatePerlinNoise1D(int noiseResolution, ReUnirand::MarsagliaUniRng &rng) {
        const int numLatticePoints = noiseResolution + 1;
        std::vector<float> gradients(numLatticePoints);

        // Generate gradients at lattice points (mapped from [0,1) to [-1,1]).
        for (int i = 0; i < numLatticePoints; ++i) {
            float randomValue = rng.uni();
            gradients[i] = 2.0f * randomValue - 1.0f;
        }

        std::vector<float> noise(noiseResolution);
        for (int x = 0; x < noiseResolution; ++x) {
            // Map the sample index to a lattice position.
            float xf = static_cast<float>(x) * (numLatticePoints - 1) / (noiseResolution - 1);
            int i0 = static_cast<int>(std::floor(xf));
            int i1 = i0 + 1;
            if (i1 >= numLatticePoints)
                i1 = numLatticePoints - 1;

            float t = xf - i0;
            float u = fade(t);

            // Calculate contributions from the two lattice points.
            float d0 = gradients[i0] * t;
            float d1 = gradients[i1] * (t - 1.0f);

            // Interpolate between the two contributions.
            noise[x] = lerp(d0, d1, u);
        }
        return noise;
    }

    std::vector<float> generatePerlinNoise2D(int noiseWidth, int noiseHeight, ReUnirand::MarsagliaUniRng &rng) {
        const float PI = 3.14159265f;
        // The grid has one extra row and column.
        int gridWidth = noiseWidth + 1;
        int gridHeight = noiseHeight + 1;

        // Define a simple 2D vector structure for internal use.
        struct Vec2 {
            float x, y;
        };

        // Generate a grid of random unit gradient vectors.
        std::vector<Vec2> gradients(gridWidth * gridHeight);
        for (int j = 0; j < gridHeight; ++j) {
            for (int i = 0; i < gridWidth; ++i) {
                float angle = rng.uni() * 2.0f * PI;
                gradients[j * gridWidth + i] = { std::cos(angle), std::sin(angle) };
            }
        }

        // Prepare the output noise array.
        std::vector<float> noise(noiseWidth * noiseHeight, 0.0f);

        // For each sample in the output noise image:
        // We define the continuous coordinate system so that the output spans from 0 to noiseWidth in x
        // and 0 to noiseHeight in y. (You can later add a frequency scaling factor if desired.)
        for (int j = 0; j < noiseHeight; ++j) {
            for (int i = 0; i < noiseWidth; ++i) {
                // Map the discrete sample indices to continuous coordinates.
                // This maps i in [0, noiseWidth-1] to sampleX in [0, noiseWidth].
                float sampleX = static_cast<float>(i) / (noiseWidth - 1) * noiseWidth;
                float sampleY = static_cast<float>(j) / (noiseHeight - 1) * noiseHeight;

                // Determine the cell (lattice) coordinates.
                int i0 = static_cast<int>(std::floor(sampleX));
                int j0 = static_cast<int>(std::floor(sampleY));
                int i1 = i0 + 1;
                int j1 = j0 + 1;
                if (i1 >= gridWidth) i1 = gridWidth - 1;
                if (j1 >= gridHeight) j1 = gridHeight - 1;

                // Compute the fractional part within the cell.
                float fx = sampleX - i0;
                float fy = sampleY - j0;

                // Helper lambda to compute the dot product of the gradient at a lattice point with the distance vector.
                auto dotProduct = [&](int ix, int jy, float dx, float dy) -> float {
                    const Vec2 &grad = gradients[jy * gridWidth + ix];
                    return grad.x * dx + grad.y * dy;
                };

                // Compute dot products for the four corners of the cell.
                float d00 = dotProduct(i0, j0, fx,    fy);
                float d10 = dotProduct(i1, j0, fx - 1, fy);
                float d01 = dotProduct(i0, j1, fx,    fy - 1);
                float d11 = dotProduct(i1, j1, fx - 1, fy - 1);

                // Apply the fade function to the fractional parts.
                float u = fade(fx);
                float v = fade(fy);

                // Interpolate along x for the top and bottom edges.
                float nx0 = lerp(d00, d10, u);
                float nx1 = lerp(d01, d11, u);

                // Interpolate along y between the two results.
                float nvalue = lerp(nx0, nx1, v);

                noise[j * noiseWidth + i] = nvalue;
            }
        }

        return noise;
    }

    std::vector<float> smoothNoise1D(const std::vector<float> &noise, int kernelRadius) {
        int n = static_cast<int>(noise.size());
        std::vector<float> smoothed(n, 0.0f);
        int kernelSize = 2 * kernelRadius + 1;
        std::vector<float> kernel(kernelSize, 0.0f);

        // Define the kernel using a parabolic (quadratic) weight function.
        float d = static_cast<float>(kernelRadius + 1);
        float sum = 0.0f;
        for (int i = -kernelRadius; i <= kernelRadius; ++i) {
            float weight = std::max(0.0f, 1.0f - (i * i) / (d * d));
            kernel[i + kernelRadius] = weight;
            sum += weight;
        }

        // Normalise the kernel weights.
        for (auto &w : kernel) {
            w /= sum;
        }

        // Convolve the noise with the kernel.
        for (int i = 0; i < n; ++i) {
            float accum = 0.0f;
            for (int j = -kernelRadius; j <= kernelRadius; ++j) {
                int index = i + j;
                // Clamp the index to the valid range.
                if (index < 0) index = 0;
                if (index >= n) index = n - 1;
                accum += noise[index] * kernel[j + kernelRadius];
            }
            smoothed[i] = accum;
        }
        return smoothed;
    }

    std::vector<float> smoothNoise2D(const std::vector<float>& noise, int noiseWidth, int noiseHeight, int kernelRadius) {
    // Build a 1D parabolic kernel.
        int kernelSize = 2 * kernelRadius + 1;
        std::vector<float> kernel(kernelSize, 0.0f);
        float d = static_cast<float>(kernelRadius + 1);
        float sum = 0.0f;
        for (int i = -kernelRadius; i <= kernelRadius; ++i) {
            float weight = std::max(0.0f, 1.0f - (i * i) / (d * d));
            kernel[i + kernelRadius] = weight;
            sum += weight;
        }
        // Normalize the kernel.
        for (auto &w : kernel) {
            w /= sum;
        }

        // Horizontal convolution.
        std::vector<float> horiz(noiseWidth * noiseHeight, 0.0f);
        for (int j = 0; j < noiseHeight; ++j) {
            for (int i = 0; i < noiseWidth; ++i) {
                float accum = 0.0f;
                for (int k = -kernelRadius; k <= kernelRadius; ++k) {
                    int idx = i + k;
                    if (idx < 0) idx = 0;
                    if (idx >= noiseWidth) idx = noiseWidth - 1;
                    accum += noise[j * noiseWidth + idx] * kernel[k + kernelRadius];
                }
                horiz[j * noiseWidth + i] = accum;
            }
        }

        // Vertical convolution.
        std::vector<float> result(noiseWidth * noiseHeight, 0.0f);
        for (int j = 0; j < noiseHeight; ++j) {
            for (int i = 0; i < noiseWidth; ++i) {
                float accum = 0.0f;
                for (int k = -kernelRadius; k <= kernelRadius; ++k) {
                    int idx = j + k;
                    if (idx < 0) idx = 0;
                    if (idx >= noiseHeight) idx = noiseHeight - 1;
                    accum += horiz[idx * noiseWidth + i] * kernel[k + kernelRadius];
                }
                result[j * noiseWidth + i] = accum;
            }
        }

        return result;
    }

} // namespace ReNoise
